// Server.cs
// Copyright 2008, Martin Green <martin@martsoft.co.uk>

/*  This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

using System;
using System.IO;
using System.Net;
using System.Net.Sockets;
using System.Threading;

namespace Kml
{
    class Server
    {
        private HttpListener listener;
        private Thread listenThread;
        private bool running;
        
        // the information to be inserted into the kml file
        private const string name = "Scarborough Castle";
        private const string description = "This is a Castle, in Scarborough.<br />Link to Wikipedia: <a href=\"http://en.wikipedia.org/wiki/Scarborough_Castle\">Scarborough Castle</a><br />";

        public Server()
        {
            // create a new Http Listener
            listener = new HttpListener();

            // ensure users can access anonymously
            listener.AuthenticationSchemes = AuthenticationSchemes.Anonymous;
            
            // add the ip:ports to listen on
            listener.Prefixes.Add("http://localhost:81/kml/");
        }

        public void start()
        {
            // start the listener and the thread
            try {
                listener.Start();
            } catch (SocketException) {
                Console.WriteLine("Cannot Listen on Port. May be in use or not running as root/Administrator");
                Environment.Exit(1);
            }
            running = true;
            // NOTE: using the listener and a thread in this way can only serve one request at a time and may not stop nicely
            listenThread = new Thread(this.listenLoop);
            listenThread.Start();
        }

        public void stop()
        {
            // stop the listener
            running = false;
            listener.Stop();
        }

        void listenLoop()
        {
            // this runs in a seperate thread, repeatedly asking the listener for a connection and reponding to it
            while (running)
            {
                try
                {
                    // get the connection context, blocks until browser connects
                    HttpListenerContext ctx = listener.GetContext();
                    // set the response code to 200 OK
                    ctx.Response.StatusCode = 200;

                    // use the request object to find out what the browser wants
                    if (ctx.Request.Url.AbsoluteUri.EndsWith("icon.png"))
                    {
                        // url ends with icon.png so return the image
                        
                        // set content type on the response
                        ctx.Response.ContentType = "image/png";
                        
                        // open the image file and copy all data to the response's output stream
                        CopyStream(File.OpenRead("icon.png"), ctx.Response.OutputStream);

                        // flush the output stream just to be sure it has sent the data
                        ctx.Response.OutputStream.Flush();
                    }
                    else if (ctx.Request.Url.AbsoluteUri.EndsWith("example.kml"))
                    {
                        // url ends with example.kml so process and return this file
                        
                        ctx.Response.ContentType = "text/xml"; // set content to XML
                        
                        // open the kml file and create in/out writers
                        StreamWriter writer = new StreamWriter(ctx.Response.OutputStream);
                        StreamReader read = new StreamReader(File.OpenRead("example.kml"));
                        
                        while (!read.EndOfStream)
                        {
                            // read a line from the source file
                            string data = read.ReadLine();

                            // replace any special tags with the required value
                            if (data.Contains("<!NAME!>"))
                                data = data.Replace("<!NAME!>", name);
                            if (data.Contains("<!DESC!>"))
                                data = data.Replace("<!DESC!>", name + "<br />" + description + "<br />Browser: "
                                    + ctx.Request.UserAgent + "<br />Time: " + DateTime.Now.ToLongTimeString());
                            if (data.Contains("<!ICON!>"))
                                data = data.Replace("<!ICON!>", "icon.png");
                            
                            // and write it out
                            writer.WriteLine(data);
                        }
                        read.Close();
                        writer.Flush();
                    }

                    // close the response, then loop to start listening for another
                    ctx.Response.Close();
                }
                catch (Exception ex)
                {
                    Console.WriteLine(ex.Message);
                }
            }
        }
        
        // simple function to copy all bytes from one stream to another
        public static void CopyStream(Stream source, Stream dest)
        {
            byte[] buffer = new byte[1024];
            int read;
            do
            {
                read = source.Read(buffer, 0, buffer.Length);
                dest.Write(buffer, 0, read);
            } while (read > 0);
        }

    }
}
